==> atmsupport-1.0.tar.gz: atmsupport-1.0/libats/ats_exp.c <==
/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libats/ats_exp.c,v 1.10 2003/09/25 09:19:53 hbb Exp $
 */

#include <stdlib.h>
#include <inttypes.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <err.h>

#include "ats.h"

static intmax_t expr(const char **str);

#define	E(L, ...)	do {				\
	if (ISVERB('e')) {				\
		printf("%s(%u): ", __func__, (L));	\
		printf(__VA_ARGS__);			\
		printf("\n");				\
	}						\
    } while (0)

static void
skip_space(const char **str)
{
	while (**str != '\0' && isascii(**str) && isspace(**str))
		(*str)++;
}

static intmax_t
num(const char **str)
{
	intmax_t a;
	char *end;

	E(0, "'%s'", *str);

	skip_space(str);
	errno = 0;
	a = strtoumax(*str, &end, 0);
	if (errno != 0 || end == *str)
		err_file("error in integer");
	if (a > INTMAX_MAX)
		err_file("integer out of range");
	*str = end;
	return (a);
}

static intmax_t
prim(const char **str)
{
	intmax_t a;
	const char *p;

	E(0, "'%s'", *str);

	skip_space(str);
	switch (**str) {

	  case '(':
		(*str)++;
		a = expr(str);
		skip_space(str);
		if (**str != ')')
			err_file("')' expected: '%s'", *str);
		(*str)++;
		break;

	  case '+':
		(*str)++;
		a = prim(str);;
		break;

	  case '-':
		(*str)++;
		a = -prim(str);;
		break;

	  case '~':
		(*str)++;
		a = !prim(str);;
		break;

	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
		a = num(str);
		break;

	  default:
		if (isascii(**str) && (isalpha(**str) || **str == '_')) {
			p = var_subst(*str, str);
			a = num(&p);
			if (*p != '\0')
				err_file("variable has non-numeric value '%s'",
				    p);
			break;
		}
		err_file("error in expression: '%s'", *str);
	}
	return (a);
}

static intmax_t
term(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);

	a = prim(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	switch (**str) {

	  case '*':
		(*str)++;
		a *= prim(str);
		goto again;

	  case '/':
		(*str)++;
		a /= prim(str);
		goto again;

	  case '%':
		(*str)++;
		a %= prim(str);
		goto again;
	}
	return (a);

}

static intmax_t
add(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);

	a = term(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	switch (**str) {

	  case '+':
		(*str)++;
		a += term(str);
		goto again;

	  case '-':
		(*str)++;
		a -= term(str);
		goto again;
	}
	return (a);
}

static intmax_t
shift(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);

	a = add(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	switch (**str) {

	  case '>':
		if ((*str)[1] == '>') {
			(*str) += 2;
			a >>= add(str);
			goto again;
		}
		break;

	  case '<':
		if ((*str)[1] == '<') {
			(*str) += 2;
			a <<= add(str);
			goto again;
		}
		break;

	}
	return (a);
}

static intmax_t
cmp(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);
	a = shift(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '<') {
		if ((*str)[1] == '=') {
			(*str) += 2;
			a = (a <= shift(str));
		} else {
			(*str)++;
			a = (a < shift(str));
		}
		goto again;
	}
	if ((*str)[0] == '>') {
		if ((*str)[1] == '=') {
			(*str) += 2;
			a = (a >= shift(str));
		} else {
			(*str)++;
			a = (a > shift(str));
		}
		goto again;
	}
	return (a);
}

static intmax_t
eq(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);
	a = cmp(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '=' && (*str)[1] == '=') {
		(*str) += 2;
		a = (a == cmp(str));
		goto again;
	}
	if ((*str)[0] == '!' && (*str)[1] == '=') {
		(*str) += 2;
		a = (a != cmp(str));
		goto again;
	}
	return (a);
}

static intmax_t
and(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);
	a = eq(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '&' && (*str)[1] != '&') {
		(*str)++;
		a &= eq(str);
		goto again;
	}
	return (a);
}

static intmax_t
andand(const char **str)
{
	intmax_t a, b;

	E(0, "'%s'", *str);
	a = and(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '&' && (*str)[1] == '&') {
		(*str) += 2;
		b = and(str);
		a = a && b;
		goto again;
	}
	return (a);
}

static intmax_t
oror(const char **str)
{
	intmax_t a, b;

	E(0, "'%s'", *str);
	a = andand(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '|' && (*str)[1] == '|') {
		(*str) += 2;
		b = andand(str);
		a = a || b;
		goto again;
	}
	return (a);
}

static intmax_t
quest(const char **str)
{
	intmax_t a, b, c;

	E(0, "'%s'", *str);
	a = oror(str);
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '?') {
		(*str)++;
		b = quest(str);
		E(2, "'%s'", *str);
		skip_space(str);
		if ((*str)[0] != ':')
			err_file("':' expected");
		(*str)++;
		c = quest(str);
		return (a ? b : c);
	}
	return (a);
}

static intmax_t
expr(const char **str)
{
	intmax_t a;
	struct var *v;
	const char *val, *end, *end_val, **save;
	int c;
	char sbuf[100];

	E(0, "'%s'", *str);
	skip_space(str);
	save = str;
	if (isascii(**str) && isalpha(**str)) {
		v = var_find(*str, &end);
		skip_space(&end);
		switch (c = *end) {
		  case '+':
		  case '-':
		  case '*':
		  case '/':
			if (end[1] != '=')
				goto noass;
			*str = end + 2;
			val = var_getval(v);
			end_val = val;
			skip_space(&end_val);
			a = num(&end_val);
			skip_space(&end_val);
			if (*end_val != '\0')
				err_file("variable has non-numeric value '%s'",
				    val);
			switch (c) {
			  case '+':
				a += expr(str);
				break;
			  case '-':
				a -= expr(str);
				break;
			  case '*':
				a *= expr(str);
				break;
			  case '/':
				a /= expr(str);
				break;
			}
			break;

		  case '=':
			if (end[1] == '=')
				goto noass;
			*str = end + 1;
			a = expr(str);
			break;

		  default:
			goto noass;
		}
		sprintf(sbuf, "%jd", a);
		var_setn(v, sbuf, strlen(sbuf));
		return (a);
	}

  noass:
	str = save;
	E(1, "'%s'", *str);
	return (quest(str));
}

/*
 * Make expression substitutions on the given line starting at ptr.
 */
char *
expr_subst(const char *ptr, const char **end)
{
	intmax_t e;
	static char sbuf[100];

	if (ISVERB('e'))
		printf("VERB: %s: '%s'\n", __func__, ptr);

	if (*ptr != '(')
		err_file("'(' expected");
	*end = ++ptr;
	e = expr(end);
	if (**end != ')')
		err_file("')' expected");
	(*end)++;
	sprintf(sbuf, "%jd", e);

	if (ISVERB('e'))
		printf("VERB: replace '%.*s' -> '%s'\n",
		    (int)(*end - ptr), ptr, sbuf);

	return (sbuf);
}

==> atmsupport-1.4.tar.gz: atmsupport-1.4/libats/ats_exp.c <==
/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libats/ats_exp.c,v 1.11 2004/07/08 13:33:34 brandt Exp $
 */

#include <stdlib.h>
#include <inttypes.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include "ats.h"
#include "atspriv.h"

static intmax_t expr(const char **str);

#if defined(__GNUC__) && __GNUC__ < 3
#define	E(L, A...)	do {				\
	if (ISVERB('e')) {				\
		printf("%s(%u): ", __FUNCTION__, (L));	\
		printf(A);			\
		printf("\n");				\
	}						\
    } while (0)
#else
#define	E(L, ...)	do {				\
	if (ISVERB('e')) {				\
		printf("%s(%u): ", __func__, (L));	\
		printf(__VA_ARGS__);			\
		printf("\n");				\
	}						\
    } while (0)
#endif

static void
skip_space(const char **str)
{
	while (**str != '\0' && isascii(**str) && isspace(**str))
		(*str)++;
}

static intmax_t
num(const char **str)
{
	intmax_t a;
	char *end;

	E(0, "'%s'", *str);

	skip_space(str);
	errno = 0;
	a = strtoumax(*str, &end, 0);
	if (errno != 0 || end == *str)
		err_file("error in integer");
	if (a > INTMAX_MAX)
		err_file("integer out of range");
	*str = end;
	return (a);
}

static intmax_t
prim(const char **str)
{
	intmax_t a;
	const char *p;

	E(0, "'%s'", *str);

	skip_space(str);
	switch (**str) {

	  case '(':
		(*str)++;
		a = expr(str);
		skip_space(str);
		if (**str != ')')
			err_file("')' expected: '%s'", *str);
		(*str)++;
		break;

	  case '+':
		(*str)++;
		a = prim(str);;
		break;

	  case '-':
		(*str)++;
		a = -prim(str);;
		break;

	  case '~':
		(*str)++;
		a = !prim(str);;
		break;

	  case '0': case '1': case '2': case '3': case '4':
	  case '5': case '6': case '7': case '8': case '9':
		a = num(str);
		break;

	  default:
		if (isascii(**str) && (isalpha(**str) || **str == '_')) {
			p = var_subst(*str, str);
			a = num(&p);
			if (*p != '\0')
				err_file("variable has non-numeric value '%s'",
				    p);
			break;
		}
		err_file("error in expression: '%s'", *str);
	}
	return (a);
}

static intmax_t
term(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);

	a = prim(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	switch (**str) {

	  case '*':
		(*str)++;
		a *= prim(str);
		goto again;

	  case '/':
		(*str)++;
		a /= prim(str);
		goto again;

	  case '%':
		(*str)++;
		a %= prim(str);
		goto again;
	}
	return (a);

}

static intmax_t
add(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);

	a = term(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	switch (**str) {

	  case '+':
		(*str)++;
		a += term(str);
		goto again;

	  case '-':
		(*str)++;
		a -= term(str);
		goto again;
	}
	return (a);
}

static intmax_t
shift(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);

	a = add(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	switch (**str) {

	  case '>':
		if ((*str)[1] == '>') {
			(*str) += 2;
			a >>= add(str);
			goto again;
		}
		break;

	  case '<':
		if ((*str)[1] == '<') {
			(*str) += 2;
			a <<= add(str);
			goto again;
		}
		break;

	}
	return (a);
}

static intmax_t
cmp(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);
	a = shift(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '<') {
		if ((*str)[1] == '=') {
			(*str) += 2;
			a = (a <= shift(str));
		} else {
			(*str)++;
			a = (a < shift(str));
		}
		goto again;
	}
	if ((*str)[0] == '>') {
		if ((*str)[1] == '=') {
			(*str) += 2;
			a = (a >= shift(str));
		} else {
			(*str)++;
			a = (a > shift(str));
		}
		goto again;
	}
	return (a);
}

static intmax_t
eq(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);
	a = cmp(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '=' && (*str)[1] == '=') {
		(*str) += 2;
		a = (a == cmp(str));
		goto again;
	}
	if ((*str)[0] == '!' && (*str)[1] == '=') {
		(*str) += 2;
		a = (a != cmp(str));
		goto again;
	}
	return (a);
}

static intmax_t
and(const char **str)
{
	intmax_t a;

	E(0, "'%s'", *str);
	a = eq(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '&' && (*str)[1] != '&') {
		(*str)++;
		a &= eq(str);
		goto again;
	}
	return (a);
}

static intmax_t
andand(const char **str)
{
	intmax_t a, b;

	E(0, "'%s'", *str);
	a = and(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '&' && (*str)[1] == '&') {
		(*str) += 2;
		b = and(str);
		a = a && b;
		goto again;
	}
	return (a);
}

static intmax_t
oror(const char **str)
{
	intmax_t a, b;

	E(0, "'%s'", *str);
	a = andand(str);
  again:
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '|' && (*str)[1] == '|') {
		(*str) += 2;
		b = andand(str);
		a = a || b;
		goto again;
	}
	return (a);
}

static intmax_t
quest(const char **str)
{
	intmax_t a, b, c;

	E(0, "'%s'", *str);
	a = oror(str);
	E(1, "'%s'", *str);
	skip_space(str);
	if ((*str)[0] == '?') {
		(*str)++;
		b = quest(str);
		E(2, "'%s'", *str);
		skip_space(str);
		if ((*str)[0] != ':')
			err_file("':' expected");
		(*str)++;
		c = quest(str);
		return (a ? b : c);
	}
	return (a);
}

static intmax_t
expr(const char **str)
{
	intmax_t a;
	struct var *v;
	const char *val, *end, *end_val, **save;
	int c;
	char sbuf[100];

	E(0, "'%s'", *str);
	skip_space(str);
	save = str;
	if (isascii(**str) && isalpha(**str)) {
		v = var_find(*str, &end);
		skip_space(&end);
		switch (c = *end) {
		  case '+':
		  case '-':
		  case '*':
		  case '/':
			if (end[1] != '=')
				goto noass;
			*str = end + 2;
			val = var_getval(v);
			end_val = val;
			skip_space(&end_val);
			a = num(&end_val);
			skip_space(&end_val);
			if (*end_val != '\0')
				err_file("variable has non-numeric value '%s'",
				    val);
			switch (c) {
			  case '+':
				a += expr(str);
				break;
			  case '-':
				a -= expr(str);
				break;
			  case '*':
				a *= expr(str);
				break;
			  case '/':
				a /= expr(str);
				break;
			}
			break;

		  case '=':
			if (end[1] == '=')
				goto noass;
			*str = end + 1;
			a = expr(str);
			break;

		  default:
			goto noass;
		}
		sprintf(sbuf, "%"FMT_INTMAX"d", a);
		var_setn(v, sbuf, strlen(sbuf));
		return (a);
	}

  noass:
	str = save;
	E(1, "'%s'", *str);
	return (quest(str));
}

/*
 * Make expression substitutions on the given line starting at ptr.
 */
char *
expr_subst(const char *ptr, const char **end)
{
	intmax_t e;
	static char sbuf[100];

	E(0, "'%s'\n", ptr);

	if (*ptr != '(')
		err_file("'(' expected");
	*end = ++ptr;
	e = expr(end);
	if (**end != ')')
		err_file("')' expected");
	(*end)++;
	sprintf(sbuf, "%"FMT_INTMAX"d", e);

	E(0, "replace '%.*s' -> '%s'\n", (int)(*end - ptr), ptr, sbuf);

	return (sbuf);
}
